/*
NAME
	proxy-via-http.c

SYNOPSYS
	int
	proxy_via_http (register FILE *const xfile,
	                register char const *const id,
	                register char const *const proxied_host_name,
	                register unsigned const proxied_port,
	                register unsigned const io_timeout,
	                register Bool const pverbose);
	
	int
	proxy_via_http_2 (register FILE *const xfile,
	                  register char const *const id,
	                  register char const *const proxied_host_name,
	                  register unsigned const proxied_port,
	                  register unsigned const io_timeout,
	                  register Bool const pverbose);

DESCRIPTION
	Given an already opened stdio file handle that has already been
	connected to some HTTP proxy server, use the HTTP CONNECT or the
	variant HTTP CONNECT protocol to connect through the proxy to some
	other server elsewhere.

AUTHOR
	Ronald F. Guilmette <rfg@monkeys.com>

COPYRIGHT
	Copyright (c) 2002 Ronald F. Guilmette; All rights reserved.

	The author and copyright holder of this software specifically and
	explicitly DISALLOWS any and all uses of this software, or any
	compiled and/or derivative version thereof for any and all commercial
	purposes without the express written permission of the copyright
	holder to the contrary.

	Requests for permission to use this software, or any compiled and/or
	derivative work thereof, either in a commercial context or for com-
	mercial gain or for any commercial purpose should be e-mailed to the
	copyright holder at the e-mail address given above.

DISCLAIMER
	THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
	TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
	PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
	OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
	LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
	USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
	AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
	OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
	OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
	SUCH DAMAGE.
*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>  /* For MAXDNAME */

#include "common.h"
#include "errors.h"
#include "utils.h"
#include "proxy-via-http.h"

typedef struct in_addr in_addr;

enum { line_max = 1 << 10 };

static char const content_length_hdr[] = "Content-length:";
enum { content_length_hdr_len = sizeof content_length_hdr - 1 };
static char const content_type_hdr[] = "Content-type:";
enum { content_type_hdr_len = sizeof content_type_hdr - 1 };
#if 0
static char const conn_close_hdr[] = "Connection: close";
enum { conn_close_hdr_len = sizeof conn_close_hdr - 1 };
#endif
static char const proxy_agent_hdr[] = "Proxy-agent:";
enum { proxy_agent_hdr_len = sizeof proxy_agent_hdr - 1 };

static size_t
readline (register char *const line,
          register size_t const line_max,
          register FILE *const xfile,
          register unsigned const io_timeout,
          register char const *const id)
{
  register size_t r;
    
  if (!(r = t_xfgets (line, line_max, xfile, io_timeout)))
    {
      if (ferror (xfile))
        notice ("%s: HCONNECT: Read error: %s", id, strerror (errno));
      else if (feof (xfile))
        notice ("%s: HCONNECT: EOF while reading", id);
      else
        notice ("%s: HCONNECT: Line too long", id);
      fclose (xfile);
      return 0;
    }

  return r;
}

static int
put_command (register FILE *const xfile,
             register char const *const id,
             register unsigned long const io_timeout,
             register char const *const command,
             register Bool const protocol_verbose)
{
  register size_t len = strlen (command);
  register size_t r;

  if (protocol_verbose)
    info ("<<< %.*s", (int)(len - 2), command);

  if ((r = t_fwrite (command, 1, len, xfile, io_timeout)) != len)
    {
      if (ferror (xfile))
        notice ("%s: HCONNECT: Write error: %s", id, strerror (errno));
      else if (feof (xfile))
        notice ("%s: HCONNECT: EOF while writing", id);
      else
        notice ("%s: HCONNECT: Short write: wrote %u, expected %u",
		id, r, len);
      fclose (xfile);
      return -1;
    }
  return 0;
}

static int
get_http_status (register char const *const line,
                 register long const count)
{
  register char const *const limit = line + count;  
  register char const *q;
  
  for (q = line; q < limit && !isspace (*q); q++)
    continue;
  for (; q < limit && isspace (*q); q++)
    continue;
  if ((q + 3) >= limit)
    return -1;
  if (!isdigit (q[0]) || !isdigit (q[1]) || !isdigit (q[2]))
    return -1;
  if (q[0] == '0')
    return -1;
  return ((q[0] - '0') * 100) + ((q[1] - '0') * 10) + (q[2] - '0');
}

int
proxy_via_http (register FILE *const xfile,
	        register char const *const id,
	        register char const *const proxied_host_name,
	        register unsigned const proxied_port,
	        register unsigned const io_timeout,
	        register Bool const pverbose)
{
  auto char line[line_max+1];
  register int r;
  register int http_status = -1;
  register Bool first_line;
  register Bool had_proxy_content_length = False;
  register Bool had_proxy_content_type = False;
#if 0
  register Bool had_conn_close = False;
#endif

  sprintf (line, "CONNECT %s:%d HTTP/1.0\r\n", proxied_host_name, proxied_port);
  if (put_command (xfile, id, io_timeout, line, pverbose) == -1)
    return -1;

  if (put_command (xfile, id, io_timeout, "\r\n", pverbose) == -1)
    return -1;

  /* Read all of the http response headers.  */
  for (first_line = True;; first_line = False)
    {
      if (!(r = readline (line, line_max, xfile, io_timeout, id)))
        return -1;
      if (pverbose)
        showline (line, r);
      if (first_line)
	http_status = get_http_status (line, r);
      else if (!strncasecmp (line, content_length_hdr, content_length_hdr_len))
	had_proxy_content_length = True;
      else if (!strncasecmp (line, content_type_hdr, content_type_hdr_len))
	had_proxy_content_type = True;
#if 0
      else if (!strncasecmp (line, conn_close_hdr, conn_close_hdr_len))
	had_conn_close = True;
#endif
      if (r == 2 && line[0] == '\r' && line[1] == '\n')
	break;  /* End of headers.  */

      /* Special kludge, mostly for AnalogX drain-bamage. */
      /* NOTE:  This protocol pecularity has also been seen to affect
	 the following additional HTTP proxy types:

		Apache/1.3.x
		SAMBAR 4.1
		SAMBAR 4.2
		SAMBAR 4.3
		SAMBAR 4.4
		SAMBAR 5.0
		SAMBAR 5.1
		Microsoft-IIS/4.0
		IJ/2.0.x
		WebGuard
		Software602 (1.01)
		ChinaWall
		DCFW-2000
		PE
		HwaRang/1.0.0
      */
      if (r == 1 && line[0] == '\n')
	{
	  notice ("%s: Non-standard HTTP headers termination kludge", id);
	  break;  /* End of headers.  */
	}
    }

  if (http_status != 200)
    {
      notice ("%s: HTTP %d", id, http_status);
      return -1;
    }

  if (had_proxy_content_length || had_proxy_content_type)
    {
      notice ("%s: HTTP %d (bogus)", id, http_status);
      return -1;
    }

#if 0
  if (had_conn_close)
    {
      notice ("%s: HTTP %d (bogus - connection closed)", id, http_status);
      return -1;
    }
#endif

  return 0;
}

/* Just like proxy_via_http above, except that we don't send a second \r\n
   after the CONNECT command.  */

int
proxy_via_http_2 (register FILE *const xfile,
	          register char const *const id,
	          register char const *const proxied_host_name,
	          register unsigned const proxied_port,
	          register unsigned const io_timeout,
	          register Bool const pverbose)
{
  /* Versions 1.0.0 and 3.0 of the following proxy both have the same
     two problems, i.e. they only expect one \r\n pair at the end of
     the connect command, and also, they are "lazy" in returning the
     stuff that's sent back by the server we connect through to.  */
  static char const hwarang[] = "HwaRang/";
  enum { hwarang_len = sizeof hwarang - 1 };
  auto char line[line_max+1];
  register int r;
  register int http_status = -1;
  register Bool first_line;
  register Bool had_proxy_content_length = False;
  register Bool had_proxy_content_type = False;
#if 0
  register Bool had_conn_close = False;
#endif
  register Bool lazy_proxy = False;

  sprintf (line, "CONNECT %s:%d HTTP/1.0\r\n", proxied_host_name, proxied_port);
  if (put_command (xfile, id, io_timeout, line, pverbose) == -1)
    return -1;

  /* Read all of the http response headers.  */
  for (first_line = True;; first_line = False)
    {
      if (!(r = readline (line, line_max, xfile, io_timeout, id)))
        return -1;
      if (pverbose)
        showline (line, r);
      if (first_line)
	http_status = get_http_status (line, r);
      else if (!strncasecmp (line, content_length_hdr, content_length_hdr_len))
	had_proxy_content_length = True;
      else if (!strncasecmp (line, content_type_hdr, content_type_hdr_len))
	had_proxy_content_type = True;
#if 0
      else if (!strncasecmp (line, conn_close_hdr, conn_close_hdr_len))
	had_conn_close = True;
#endif
      else if (!strncasecmp (line, proxy_agent_hdr, proxy_agent_hdr_len))
	{
	  if (!strncasecmp (&line[13], hwarang, hwarang_len))
	    {
	      notice ("%s: HwaRang lazy proxy kludge", id);
	      lazy_proxy = True;
	    }
	}
      if (r == 2 && line[0] == '\r' && line[1] == '\n')
	break;  /* End of headers.  */

      /* Special kludge, mostly for AnalogX drain-bamage. */
      /* NOTE:  This protocol pecularity has also been seen to affect
	 the following additional HTTP proxy types:

		Apache/1.3.x
		SAMBAR 4.1
		SAMBAR 4.2
		SAMBAR 4.3
		SAMBAR 4.4
		SAMBAR 5.0
		SAMBAR 5.1
		Microsoft-IIS/4.0
		IJ/2.0.x
		WebGuard
		Software602 (1.01)
		ChinaWall
		DCFW-2000
		PE
		HwaRang/1.0.0
      */
      if (r == 1 && line[0] == '\n')
	{
	  notice ("%s: Non-standard HTTP headers termination kludge", id);
	  break;  /* End of headers.  */
	}
    }

  if (http_status != 200)
    {
      notice ("%s: HTTP %d", id, http_status);
      return -1;
    }

  if (had_proxy_content_length || had_proxy_content_type)
    {
      notice ("%s: HTTP %d (bogus)", id, http_status);
      return -1;
    }

#if 0
  if (had_conn_close)
    {
      notice ("%s: HTTP %d (bogus - connection closed)", id, http_status);
      return -1;
    }
#endif

  return lazy_proxy ? 1 : 0;
}
