/*
NAME
	proxy-via-socks.c

SYNOPSYS
	int
	proxy_via_socks5 (register FILE *const xfile,
	                  register char const *const id,
	                  register char const *const proxied_host_name,
	                  register unsigned const proxied_port,
	                  register unsigned const io_timeout,
	                  register Bool const pverbose);

	int
	proxy_via_socks4 (register FILE *const xfile,
	                  register char const *const id,
	                  register char const *const proxied_host_name,
	                  register unsigned const proxied_port,
	                  register unsigned const io_timeout,
	                  register Bool const pverbose);

DESCRIPTION

AUTHOR
	Ronald F. Guilmette <rfg@monkeys.com>

COPYRIGHT
	Copyright (c) 2002 Ronald F. Guilmette; All rights reserved.

	The author and copyright holder of this software specifically and
	explicitly DISALLOWS any and all uses of this software, or any
	compiled and/or derivative version thereof for any and all commercial
	purposes without the express written permission of the copyright
	holder to the contrary.

	Requests for permission to use this software, or any compiled and/or
	derivative work thereof, either in a commercial context or for com-
	mercial gain or for any commercial purpose should be e-mailed to the
	copyright holder at the e-mail address given above.

DISCLAIMER
	THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
	TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
	PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
	OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
	LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
	USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
	AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
	OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
	OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
	SUCH DAMAGE.
*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>  /* For MAXDNAME */
#include <netdb.h>

#include "common.h"
#include "errors.h"
#include "utils.h"
#include "proxy-via-socks.h"

typedef struct in_addr in_addr;
typedef struct hostent hostent;

static int
type_of (register char const *const proxied_host_name)
{
  auto in_addr addr;

  if (my_inet_aton (proxied_host_name, &addr) != 0)
    return 1;
  else
    return 3;
}

static in_addr
addr_of (register char const *const proxied_host_name)
{
  auto in_addr addr;

  if (my_inet_aton (proxied_host_name, &addr) == 0)
    {
      register hostent *hp;

      if (!(hp = gethostbyname (proxied_host_name)))
	{
	  addr.s_addr = 0;
	  return addr;
	}
      memcpy (&addr, hp->h_addr, sizeof addr);
    }

  return addr;
}

int
proxy_via_socks5 (register FILE *const xfile,
	          register char const *const id,
	          register char const *const proxied_host_name,
	          register unsigned const proxied_port,
	          register unsigned const io_timeout,
	          register Bool const pverbose)
{
  enum { line_max = 1 << 10 };
  auto char buf[line_max+1];
  auto unsigned short port_in_nbo;
  auto in_addr bind_addr;
  auto char bind_dname[MAXDNAME+2];
  auto unsigned short bind_port;
  register unsigned peer_socks_version;
  register unsigned peer_acceptable_method;
  register size_t r;
  register char *p; 
  register unsigned addr_len;
  register size_t len;
  register unsigned dlen;

  buf[0] = 5;
  buf[1] = 1;
  buf[2] = 0;
  len = 3;

  if (pverbose)
    debug ("<<< {%u, %u, %u} (SOCKS5)",
           (unsigned char) buf[0],
           (unsigned char) buf[1],
           (unsigned char) buf[2]);

  if ((r = t_fwrite (buf, 1, len, xfile, io_timeout)) != len)
    {
      if (ferror (xfile))
	notice ("%s: SOCKS5: Phase 1: Write error: %s", id, strerror (errno));
      else if (feof (xfile))
	notice ("%s: SOCKS5: Phase 1: EOF while writing", id);
      else
	notice ("%s: SOCKS5: Phase 1: Short write: wrote %u, expected %u",
		id, r, len);
      return -1;
    }

  len = 2;
  if ((r = t_fread (buf, 1, len, xfile, io_timeout)) != len)
    {
      if (ferror (xfile))
        notice ("%s: SOCKS5: Phase 1: Read error: %s", id, strerror (errno));
      else if (feof (xfile))
        notice ("%s: SOCKS5: Phase 1: EOF while reading", id);
      else
        notice ("%s: SOCKS5: Phase 1: Short read: read %u, expected %u",
		id, r, len);
      return -1;
    }

  if (pverbose)
    debug (">>> {%u, %u} (SOCKS5)",
           (unsigned char) buf[0],
           (unsigned char) buf[1]);

  peer_socks_version = (unsigned char) buf[0];
  /* NOTE: RFC 1928 nver actually comes out and says what the version number
     value in the initial response from a SOCKS5 server should be.  Given
     that the older SOCKS4 protocol document says that the server response
     in that protocol should give back zero as the protocol number, some
     implementors of SOCKS5 may have felt that that was the correct value
     to put in the initial server response for the SOCKS5 protocol also.
     But some others may have assumed that the version value should be 5,
     so here we will accept either 0 or 5.  */
  if (peer_socks_version != 0 && peer_socks_version != 5)
    {
      warn ("%s: SOCKS5: Server gave invalid SOCKS version: %u",
	    id, peer_socks_version);
    }

  peer_acceptable_method = (unsigned char) buf[1];
  if (peer_acceptable_method != 0)
    {
      notice ("%s: SOCKS5: Must authenticate", id);
      return -1;
    }

  /* Build the CONNECT request.  */

  buf[0] = 5;
  buf[1] = 1;  /* CONNECT */
  buf[2] = 0;  /* reserved */
  buf[3] = type_of (proxied_host_name);
  p = &buf[4];
  if (type_of (proxied_host_name) == 1)
    {
      auto in_addr const addr = addr_of (proxied_host_name);

      addr_len = sizeof addr;
      memcpy (p, &addr, addr_len);
    }
  else
    {
      addr_len = strlen (proxied_host_name);
      *p++ = addr_len;
      memcpy (p, proxied_host_name, addr_len);
      p--;
      addr_len++;
    }
  p += addr_len;
  port_in_nbo = (unsigned short) htons (proxied_port);
  memcpy (p, &port_in_nbo, sizeof (unsigned short));
  p += sizeof (unsigned short);

  /* Write request to proxy and get response.  */

  if (pverbose)
    debug ("<<< {%u, %u, %u, %u, %s, %u} (SOCKS5)",
           (unsigned char) buf[0],
           (unsigned char) buf[1],
           (unsigned char) buf[2],
           (unsigned char) buf[3],
           proxied_host_name,
           proxied_port);

  len = p - buf;
  if ((r = t_fwrite (buf, 1, len, xfile, io_timeout)) != len)
    {
      if (ferror (xfile))
        notice ("%s: SOCKS5: Phase 2: Write error: %s", id, strerror (errno));
      else if (feof (xfile))
        notice ("%s: SOCKS5: Phase 2: EOF while writing", id);
      else
        notice ("%s: SOCKS5: Phase 2: Short write: wrote %u, expected %u",
		id, r, len);
      return -1;
    }

  /* Read only the first 2 bytes of the reply packet.

     NOTE: We really _should_ be able to count on being able to read at
     least 4 bytes from the server at this point, but experience shows
     that at least some brain-dead SOCKS5 server implementations will
     only send us back 3 bytes at this point and then stall.

     Because the second byte of the reply packet is used to contain the
     server's returned error code (if any), we can just read far enough
     into the packet to get that, and then drop the connection if the
     server has indicated an error.  This prevents us from stalling out
     while waiting for one or two more bytes that may never come.
  */

  len = 2;
  if ((r = t_fread (buf, 1, len, xfile, io_timeout)) != len)
    {
      if (ferror (xfile))
        notice ("%s: SOCKS5: Phase 2: Read error: %s", id, strerror (errno));
      else if (feof (xfile))
        notice ("%s: SOCKS5: Phase 2: EOF while reading", id);
      else
        notice ("%s: SOCKS5: Phase 2: Short read: read %u, expected %u",
		id, r, len);
      return -1;
    }

  if (pverbose)
    debug (">>> {%u, %u} (SOCKS5)",
           (unsigned char) buf[0],
           (unsigned char) buf[1]);

  switch ((unsigned char) buf[1])
    {
    case 0:
      debug ("%s: SOCKS5: Request granted", id);
      break;

    case 1:
      notice ("%s: SOCKS5: General SOCKS server failure", id);
      return -1;

    case 2:
      notice ("%s: SOCKS5: Connection not allowed by ruleset", id);
      return -1;

    case 3:
      notice ("%s: SOCKS5: Network unreachable", id);
      return -1;

    case 4:
      notice ("%s: SOCKS5: Host unreachable", id);
      return -1;

    case 5:
      notice ("%s: SOCKS5: Connection refused", id);
      return -1;

    case 6:
      notice ("%s: SOCKS5: TTL expired", id);
      return -1;

    case 7:
      notice ("%s: SOCKS5: Command not supported", id);
      return -1;

    case 8:
      notice ("%s: SOCKS5: Address type not supported", id);
      return -1;

    default:
      notice ("%s: SOCKS5: Unknown error: %u", id, (unsigned char) buf[1]);
      return -1;
  }

  /* Now we can safely read the next two bytes of the reply packet... we
     should now be able to do so without stalling.  */

  len = 2;
  if ((r = t_fread (buf, 1, len, xfile, io_timeout)) != len)
    {
      if (ferror (xfile))
        notice ("%s: SOCKS5: Phase 3: Read error: %s", id, strerror (errno));
      else if (feof (xfile))
        notice ("%s: SOCKS5: Phase 3: EOF while reading", id);
      else
        notice ("%s: SOCKS5: Phase 3: Short read: read %u, expected %u",
                id, r, len);
      return -1;
    }

  if (pverbose)
    debug (">>> {%u, %u} (SOCKS5)",
           (unsigned char) buf[0],
           (unsigned char) buf[1]);

  switch ((unsigned char) buf[1])
    {
    case 1:
      len = 6;
      if ((r = t_fread (buf, 1, len, xfile, io_timeout)) != len)
        {
          if (ferror (xfile))
            notice("%s: SOCKS5: Phase 4: Read error: %s", id, strerror (errno));
          else if (feof (xfile))
            notice ("%s: SOCKS5: Phase 4: EOF while reading", id);
          else
            notice ("%s: SOCKS5: Phase 4: Short read: read %u, expected %u",
		    id, r, len);
          return -1;
        }
      memcpy (&bind_addr, &buf[0], sizeof (in_addr));
      memcpy (&bind_port, &buf[4], sizeof (unsigned short));
      if (pverbose)
        debug(">>> {%s %u} (SOCKS5)", inet_ntoa (bind_addr), ntohs (bind_port));
      break;

    case 3:
      /* For a domain name in the response packet, read just the domain name
	 length byte first.  Then read the following domain name.  */
      len = 1;
      if ((r = t_fread (buf, 1, len, xfile, io_timeout)) != len)
        {
          if (ferror (xfile))
            notice ("%s: SOCKS5: Phase 4: Read error: %s",
		    id, strerror (errno));
          else if (feof (xfile))
            notice ("%s: SOCKS5: Phase 4: EOF while reading", id);
          else
            notice ("%s: SOCKS5: Phase 4: Short read: read %u, expected %u",
		    id, r, len);
          return -1;
        }
      if (pverbose)
	debug (">>> {%u} (SOCKS5)", (unsigned char) buf[0]);
      dlen = (unsigned char) buf[0];
      len = dlen + 2;
      if ((r = t_fread (buf, 1, len, xfile, io_timeout)) != len)
        {
          if (ferror (xfile))
            notice("%s: SOCKS5: Phase 5: Read error: %s", id, strerror (errno));
          else if (feof (xfile))
            notice ("%s: SOCKS5: Phase 5: EOF while reading", id);
          else
            notice ("%s: SOCKS5: Phase 5: Short read: read %u, expected %u",
		    id, r, len);
          return -1;
        }
      memcpy (&bind_port, &bind_dname[dlen], sizeof (unsigned short));
      bind_dname[dlen] = '\0';
      if (pverbose)
        debug (">>> {%s %u} (SOCKS5)", bind_dname, ntohs (bind_port));
      break;

    default:
      notice ("%s: SOCKS5: Invalid address type in response: %u",
             id, (unsigned char) buf[3]);
      return -1;
    }

  return 0;
}

int
proxy_via_socks4 (register FILE *xfile,
                  register char const *const id,
                  register char const *const proxied_host_name,
                  register unsigned const proxied_port,
                  register unsigned const io_timeout,
                  register Bool const pverbose)
{
  enum { line_max = 1 << 10 };
  register char const *const socks4_userid = current_userid ();
  register size_t const socks4_userid_len = strlen (socks4_userid) + 1;
  auto char buf[line_max+1];
  auto unsigned short port_in_nbo = (unsigned short) htons (proxied_port);
  auto in_addr addr;
  register unsigned peer_socks_version;
  register unsigned peer_result;
  register size_t r;
  register char *p;
  register size_t len;

  p = buf;
  *p++ = 4;
  *p++ = 1;
  memcpy (p, &port_in_nbo, sizeof (short));
  p += sizeof (short);
  addr = addr_of (proxied_host_name);
  memcpy (p, &addr, sizeof addr);
  p += sizeof addr;
  memcpy (p, socks4_userid, socks4_userid_len);
  p += socks4_userid_len;

  if (pverbose)
    debug ("<<< {%u, %u, %s, %s} (SOCKS4)",
           4, proxied_port, inet_ntoa (addr), socks4_userid);

  len = p - buf;
  if ((r = t_fwrite (buf, 1, len, xfile, io_timeout)) != len)
    {
      if (ferror (xfile))
        notice ("%s: SOCKS4: Write error: %s", id, strerror (errno));
      else if (feof (xfile))
        notice ("%s: SOCKS4: EOF while writing", id);
      else
        notice ("%s: SOCKS4: Short write: wrote %u, expected %u", id, r, len);
      return -1;
    }

  len = 8;
  if ((r = t_fread (buf, 1, len, xfile, io_timeout)) != len)
    {
      if (ferror (xfile))
        notice ("%s: SOCKS4: Read error: %s", id, strerror (errno));
      else if (feof (xfile))
        notice ("%s: SOCKS4: EOF while reading", id);
      else
        notice ("%s: SOCKS4: Short read: read %u, expected %u", id, r, len);
      return -1;
    }

  if (pverbose)
    {
      auto unsigned short port_in_nbo_out;
      auto in_addr addr_out;

      memcpy (&port_in_nbo_out, &buf[2], sizeof (unsigned short));
      memcpy (&addr_out, &buf[4], sizeof (in_addr));
      debug (">>> {%u, %u, %u, %s} (SOCKS4)",
             (unsigned char) buf[0],
             (unsigned char) buf[1],
             ntohs (port_in_nbo_out),
             inet_ntoa (addr_out));
    }

  peer_socks_version = (unsigned char) buf[0];
  /* For SOCKS4, the version number in the reply is supposed to be zero.
     That's what the official protocol docs say anyway.  But in practice
     there seem to be some SOCKS4 servers that will return 4 here, so we
     allow for either 0 or 4 here.

     NOTE:  At least one SOCKS4 server out their yield a version number
     for `48' (for no adequately known reason), so we allow for that also.  */
  if (peer_socks_version != 0
      && peer_socks_version != 4
      && peer_socks_version != 48)
    {
      warn ("%s: SOCKS4: Server gave invalid SOCKS version: %u",
            id, peer_socks_version);
    }

  peer_result = (unsigned char) buf[1];
  switch (peer_result)
    {
    case 90:
      debug ("%s: SOCKS4: Request granted", id);
      break;  /* request granted */
    case 91:
      notice ("%s: SOCKS4: Request rejected or failed", id);
      return -1;
    case 92:
      notice ("%s: SOCKS4: No identd", id);
      return -1;
    case 93:
      notice ("%s: SOCKS4: identd result inconsistant", id);
      return -1;
    default:
      notice ("%s: SOCKS4: Unknown error: %d", id, peer_result);
      return -1;
    }

  return 0;
}
