/*
NAME
	pxstress - Stress Test a SOCKS or HTTP/CONNECT proxy

SUMMARY
	ptest [-V] [-L] [-E priority] [-S facility] host:[port[:proto]]

DESCRIPTION
	Pxstress is a stress tester for SOCK4, SOCKS5, and HTTP CONNECT
	proxies.  Given a host, port, and (optionally) a protocol desig-
	nation, pxstress will attempt to connect to the given host and
	port, and if successful, it will use the specified (or inferred)
	protocol to induce the proxy to make a connection back to itself.
	This process is then repeated, ad infinitum, or for all long as
	the relevant proxy server is able to continue making connections
	back to itse own IP address and port without returning any error.

	Each proxy server candidate specification must consist of at least
	a host name or an IPv4 dotted-quad IP address.  The host name or
	IP address may be optionally followed by a colon and then a port
	number or a service name (taken from the local /etc/services file).
	The port number or service name, when given, may then optionally
	be followed, in turn, by a colon and a special two-character
	prococol abbreviation.  (See below.)

	If a proxy host name or IP address is given, but no port number
	or service name is attached to the host specification, then the
	specified host will be tested to see if it is running any of the
	following types of proxies on any of the indicated ports:

		Protocol		port	service name
		---------------		----	-------------------
		SOCKS Version 4		1080	socks
		SOCKS Version 5		1080	socks
		HTTP/CONNECT(*)		80	http
		HTTP/CONNECT(*)		81	hosts2-ns
		HTTP/CONNECT(*)		3128	squid
		HTTP/CONNECT(*)		4480	N/A (Proxy+)
		HTTP/CONNECT(*)		6588	N/A (AnalogX)
		HTTP/CONNECT(*)		8000	N/A
		HTTP/CONNECT(*)		8080	webcache
		HTTP/CONNECT(*)		8081	tproxy

	(* Note that the HTTP/CONNECT protocol has several subtle variations,
	and these will be automatically checked for when appropriate.)

	If a port number or service name is supplied in any given proxy
	specification, then only that port/service will be tested for
	proxying capability.

	As noted above, if any given proxy specification includes a port
	number or service name part, then that may be followed by a colon
	and then a protocol abbreviation.  The available protocol abbrevia-
	tions are as follows:

		s4	- SOCKS Version 4 protocol
		s5	- SOCKS Version 5 protocol
		hc	- HTTP CONNECT method protocol
		h2	- variant HTTP CONNECT method protocol

	(NOTE:  The variant HTTP CONNECT method protocol is identical to
	the standard HTTP CONNECT method protocol except that instead of
	the CONNECT command being followed by _two_ <CR><LF> line comple-
	tion character sequences, only one <CR><LF> line completion char-
	acter sequence is used at the end of the CONNECT command.  This
	variant of the basic HTTP CONNECT protocol is fairly common.)

	If, for any given proxy specification, no protocol type is explicitly
	specified, then the protocol (or protocols) that will be tested for
	that proxy are entirely determined by the port number (or service
	name) part of the proxy specification.  The following table shows
	which protocols will be tested, by default, for different possible
	proxy port numbers:

	port number	service name	default protocols tested
	-----------	------------	------------------------
	1080		socks		SOCKSv4, SOCKSv5
	<<all others>>	<<all others>>	HTTP CONNECT and possibly also the
					variant HTTP CONNECT protocol

	If port number other than 1080 (or a service name other than "socks")
	is included in any given proxy specification, but no explicit
	protocol type is specified for that proxy, then the proxy will
	be tested to see if it will perform proxying when the HTTP CONNECT
	method protocol is used.  In such cases, if and only if this proxying
	test fails in a way that causes the target SMTP server to issue a
	5xx permanent failure response rigt after the HELO SMTP command is
	issued (usually indicating a syntax error) then ptest will also
	attempt to test the given proxy using the variant HTTP CONNECT
	protocol (which may perhaps be successful in some subset of these
	exact special cases).

	Normally, pxstress prints nothing during the test.  Progress of
	the test (and all protocol interactions) may however be monitored
	via use of the -V and -E options.  (See below.)

OPTIONS
	The -V option causes all protocol activity, both with the
	selected proxy and with the target mail server, to be printed
	to the currently selected log (either stderr or syslog).  Use
	of this option in conjunction with the -P option is NOT
	recommended.  Note that the -V option also causes the version
	number of ptest to be printed to stderr.

	The -L option causes ptest to assume that any and all HTTP
	proxies are ``lazy'', i.e. that following the establishment of
	a connection through any such proxy, responses from the final
	target server will not be sent back to ptest until ptest has
	itself sent at least one line of data to the target server
	through the proxy.  In practice, this option is never actually
	useful or necessary.  (Note that ptest automatically adapts
	into lazy HTTP proxy mode whenever it detects that it is talking
	to an "HwaRang" brand HTTP proxy.  It is not necessary to specify
	the -L option for this to occur.)

	The -E option can be used to select the verbosity of the messages
	produced by ptest and written to its error channel.  (See the -S
	option below.)  The argument for the -E option must be one of the
	following symbolic syslog priority names:

		emerg (or panic)
		alert
		crit
		err (or error)
		warn (or warning)
		notice
		info
		debug

	Symbolic names that appear lower in the list above produce more
	verbosity from ptest than names higher in the list.  The default,
	if no -E option is given, is to produce messages at the "notice"
	level or above.

	The -S option may be used to select a syslog facility to which ptest
	error, warning, and notice messages should be written.  This option
	may be useful if ptest is to be run in the background or as part of
	a cron job.  The argument of the -S option must be one of the follow-
	ing symbolic syslog facility names:

        	auth
        	authpriv
        	console
        	cron
        	daemon
        	ftp
        	kern
        	lpr
        	mail
        	mark
        	news
        	ntp
        	security
        	syslog
        	user
        	uucp
        	local0
        	local1
        	local2
        	local3
        	local4
        	local5
        	local6
        	local7

	The default, if no -S option is given, is to write all ptest error
	warning, and notice messages to stderr (and to not use syslog at all).

VERSION
	1.1 - Released June 22, 2002

AUTHOR
	Ronald F. Guilmette <rfg@monkeys.com>

COPYRIGHT
	Copyright (c) 2002 Ronald F. Guilmette; All rights reserved.

	The author and copyright holder of this software specifically and
	explicitly DISALLOWS any and all uses of this software, or any
	compiled and/or derivative version thereof for any and all commercial
	purposes without the express written permission of the copyright
	holder to the contrary.

	Requests for permission to use this software, or any compiled and/or
	derivative work thereof, either in a commercial context or for com-
	mercial gain or for any commercial purpose should be e-mailed to the
	copyright holder at the e-mail address given above.

DISCLAIMER
	THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
	TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
	PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
	OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
	LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
	USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
	AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
	OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
	OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
	SUCH DAMAGE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <netdb.h>
#include <sys/wait.h>

typedef struct in_addr in_addr;
typedef struct sockaddr sockaddr;
typedef struct sockaddr_in sockaddr_in;
typedef struct hostent hostent;
typedef struct servent servent;
typedef struct tm tm;
typedef struct timeval timeval;

#include "common.h"
#include "errors.h"
#include "utils.h"
#include "proxy-via-http.h"
#include "proxy-via-wingate.h"
#include "proxy-via-socks.h"

enum { line_max = 1 << 10 };

typedef enum {
    unknown_protocol,
    http_connect_protocol,
    /* Like HTTP/CONNECT, except don't send \r\n\r\n at the end of the
       CONNECT command.  Instead just send \r\n once.  */
    http2_connect_protocol,
    socks4_protocol,
    socks5_protocol,
    wingate_protocol
} protocol;

typedef struct {
    char const  *host_name;
    char const  *port_name;
    in_addr     addr;
    unsigned    port;
    protocol	proto;
} proxy_info;

static char const version_string[] = "1.1";

static char const *pname;

static unsigned connect_timeout = 24;
static unsigned short_io_timeout = 60 * 60;

static Bool pverbose = False;
static Bool force_lazy = False;

static Bool last_proxy_lazy = False;

static void
usage (void)
{
  fatal (
   "Usage: `%s [-V] [-L] [-E priority] [-S facility] host:[port[:proto]]'",
    pname);
}

static char const *
proto_abbrev (register protocol const proto)
{
  switch (proto)
    {
    case wingate_protocol:
      return "wg";

    case socks4_protocol:
      return "s4";

    case socks5_protocol:
      return "s5";

    case http_connect_protocol:
      return "hc";

    case http2_connect_protocol:
      return "h2";

    default:
      fatal ("proto_abbrev: Invalid parameter value");
    }
}

static int
port_number (register char const *const port_name)
{
  if (all_digits (port_name))
    return atoi (port_name);
  else
    {
      register servent *sp;

      if (!(sp = getservbyname (port_name, "tcp")))
        fatal ("Invalid service name: %s", port_name);
      return ntohs (sp->s_port);
    }
}

static in_addr
addr_of (register char const *const name)
{
  auto in_addr addr;

  if (my_inet_aton (name, &addr) == 0)
    {
      register hostent *hp;
      
      if (!(hp = gethostbyname (name)))
	{
          error ("Invalid host name: %s", name);
	  addr.s_addr = 0;
	  return addr;
	}
      memcpy (&addr, hp->h_addr, sizeof addr);
    }
  return addr;
}

static int
set_proxy (register proxy_info *const pip, register char const *const spec)
{
  register char const *colon1;
  register char const *colon2;
  register unsigned len;
  register char *p;
  register char const *port_name;

  if (!(colon1 = strchr (spec, ':')))
    colon1 = spec + strlen (spec);
  len = colon1 - spec;
  pip->host_name = p = (char *) xmalloc (len + 1);
  memcpy (p, spec, len);
  p[len] = '\0';
  pip->addr = addr_of (pip->host_name);
  if (pip->addr.s_addr == 0)
    return -1;

  if (!*colon1)
    {
      pip->port_name = "<<unspecified>>";
      pip->port = 0;
      pip->proto = unknown_protocol;
      return 0;
    }

  port_name = colon1 + 1;
  colon2 = strchr (port_name, ':');
  if (!colon2)
    colon2 = port_name + strlen (port_name);
  len = colon2 - port_name;
  pip->port_name = p = (char *) xmalloc (len + 1);
  memcpy (p, port_name, len);
  p[len] = '\0';

  pip->port = port_number (pip->port_name);

  if (!*colon2)
    pip->proto = unknown_protocol;
  else
    {
      register char const *const proto_name = colon2 + 1;

      if (!strcmp (proto_name, "s4"))
	pip->proto = socks4_protocol;
      else if (!strcmp (proto_name, "s5"))
	pip->proto = socks5_protocol;
      else if (!strcmp (proto_name, "hc"))
	pip->proto = http_connect_protocol;
      else if (!strcmp (proto_name, "h2"))
	pip->proto = http2_connect_protocol;
      else if (!strcmp (proto_name, "wg"))
	pip->proto = wingate_protocol;
      else
	{
	  error ("Invalid protocol specification: %s", spec);
	  return -1;
	}
    }
  return 0;
}

static int
proxy (register FILE *const xfile,
       register char const *const this_host_name,
       register char const *const this_port_name,
       register protocol const this_proxy_proto,
       register char const *next_host_name,
       register unsigned short const next_port)
{
  auto char id[1024];
  register int r;

  sprintf (id, "%s:%s:%s", this_host_name,
	   this_port_name, proto_abbrev (this_proxy_proto));
  debug ("Proxying via: %s to %s:%u", id, next_host_name, next_port);

  /* Special hack for target server specified as 0.0.0.0.  */
  if (!strcmp (next_host_name, "0.0.0.0"))
    next_host_name = this_host_name;

  last_proxy_lazy = False;
  switch (this_proxy_proto)
    {
    case unknown_protocol:
      fatal ("%s: Unknown proxy protocol", id);

    case wingate_protocol:
      if (proxy_via_wingate (xfile, id, next_host_name, next_port,
			     short_io_timeout, pverbose) == -1)
        return -1;
      break;

    case socks4_protocol:
      if (proxy_via_socks4 (xfile, id, next_host_name, next_port,
			    short_io_timeout, pverbose) == -1)
        return -1;
      break;

    case socks5_protocol:
      if (proxy_via_socks5 (xfile, id, next_host_name, next_port,
			    short_io_timeout, pverbose) == -1)
        return -1;
      break;

    case http_connect_protocol:
      if (proxy_via_http (xfile, id, next_host_name, next_port,
			  short_io_timeout, pverbose) == -1)
        return -1;
      break;

    case http2_connect_protocol:
      if ((r = proxy_via_http_2 (xfile, id, next_host_name, next_port,
			         short_io_timeout, pverbose)) == -1)
        return -1;
      if (r)
	last_proxy_lazy = True;
      break;
    }

  return 0;
}

static void
my_connect (register proxy_info const *const pip)
{
  auto sockaddr_in sin;
  register char const *this_host_name;
  register char const *this_port_name;
  register unsigned this_port;
  register char const *next_host_name;
  register unsigned next_port;
  register protocol this_proto;
  auto char id[1024];
  register int s;
  register FILE *xfile;

  sprintf (id, "%s:%s:%s", pip->host_name,
           pip->port_name, proto_abbrev (pip->proto));

  if ((s = socket (AF_INET, SOCK_STREAM, 0)) == -1)
    fatal ("%s: Error creating socket: %s", id, strerror (errno));

  memset (&sin, '\0', sizeof sin);
  sin.sin_family = AF_INET;
  sin.sin_addr = pip->addr;
  sin.sin_port = htons (pip->port);

  debug ("Connecting to %s", id);
  if (t_connect (s, (sockaddr *) &sin, sizeof sin, connect_timeout))
    {
      info ("%s: Error in connect: %s", id, strerror (errno));
      close (s);
      return;
    }
  debug ("Connected to %s", id);
  if ((xfile = fdopen (s, "r+")) == NULL)
    {
      error ("%s: Error in fdopen: %s", id, strerror (errno));
      close (s);
      return;
    }
#if 0
  setvbuf (xfile, NULL, _IONBF, 0);
#else
  setvbuf (xfile, NULL, _IOLBF, 0);
#endif

  this_host_name = pip->host_name;
  this_port_name = pip->port_name;
  this_proto = pip->proto;
  this_port = pip->port;

  next_host_name = this_host_name;
  next_port = port_number (this_port_name);

  while (!proxy (xfile, this_host_name, this_port_name, this_proto,
		 next_host_name, next_port))
    continue;

  fclose (xfile);
}

static void
test_proxy (register proxy_info const *const pip)
{
  my_connect (pip);
}

static void
test_proxy_protocols (register proxy_info const *const pip)
{
  auto proxy_info test_pinfo;

  if (pip->proto != unknown_protocol)
    {
      test_proxy (pip);
      return;
    }
  test_pinfo = *pip;
  switch (pip->port)
    {
    case 23:
      test_pinfo.proto = wingate_protocol;
      test_proxy (&test_pinfo);
      break;
    case 1080:
      test_pinfo.proto = socks4_protocol;
      test_proxy (&test_pinfo);
      test_pinfo.proto = socks5_protocol;
      test_proxy (&test_pinfo);
      break;
    default:
      test_pinfo.proto = http_connect_protocol;
      test_proxy (&test_pinfo);
      break;
    }
}

static void
test_proxy_ports (register proxy_info const *const pip)
{
  auto proxy_info test_pinfo;

  if (pip->port)
    {
      test_proxy_protocols (pip);
      return;
    }

  test_pinfo = *pip;

#if 0
  test_pinfo.port = 23;
  test_proxy_protocols (&test_pinfo);
#endif

  test_pinfo.port = 1080;
  test_pinfo.port_name = "1080";
  test_proxy_protocols (&test_pinfo);

  test_pinfo.port = 80;
  test_pinfo.port_name = "80";
  test_proxy_protocols (&test_pinfo);
  test_pinfo.port = 81;
  test_pinfo.port_name = "81";
  test_proxy_protocols (&test_pinfo);
  test_pinfo.port = 3128;
  test_pinfo.port_name = "3128";
  test_proxy_protocols (&test_pinfo);
  /* Port 4480 is used by the Proxy+ proxy.  */
  test_pinfo.port = 4480;
  test_pinfo.port_name = "4480";
  test_proxy_protocols (&test_pinfo);
  /* Port 6588 is used by the AnalogX proxy.  */
  test_pinfo.port = 6588;
  test_pinfo.port_name = "6588";
  test_proxy_protocols (&test_pinfo);
  test_pinfo.port = 8000;
  test_pinfo.port_name = "8000";
  test_proxy_protocols (&test_pinfo);
  test_pinfo.port = 8080;
  test_pinfo.port_name = "8080";
  test_proxy_protocols (&test_pinfo);
  test_pinfo.port = 8081;
  test_pinfo.port_name = "8081";
  test_proxy_protocols (&test_pinfo);
}

static void
process_arg (register char const *const arg)
{
  auto proxy_info pinfo;

  if (set_proxy (&pinfo, arg))
    return;

  test_proxy_ports (&pinfo);
}

int
main (int argc, char **argv)
{
  register int optch;
  register char const *syslog_facility_name = NULL;
  register char const *syslog_priority_name = "notice";  /* Default.  */

  pname = strrchr (argv[0], '/');
  pname = pname ? pname+1 : argv[0];

  errors_init (pname, syslog_facility_name, syslog_priority_name);
  while ((optch = getopt (argc, argv, "VLS:E:")) != -1)
    {
      switch (optch)
        {
        case 'V':
          pverbose = True;
          break;
  
        case 'L':
          force_lazy = True;
          break;
  
        case 'S':
          syslog_facility_name = optarg;
          if (errors_init (pname, syslog_facility_name, syslog_priority_name))
            usage ();
          break;

        case 'E':
          syslog_priority_name = optarg;
          if (errors_init (pname, syslog_facility_name, syslog_priority_name))
            usage ();
          break;

        default:
          usage ();
        }
    }

  if (pverbose)
    notice ("Version: %s", version_string);

  set_signal_handler (SIGPIPE, SIG_IGN);

  if (optind >= argc)
    usage ();

  process_arg (argv[optind]);

  return 0;
}
