/*
NAME
	utils.c

SYNOPSYS
	void *
	xmalloc (register unsigned long const size);

	char *
	xstrdup (register char const *const s);

	void
	xfree (register void *const ptr);
	
	Bool
	all_digits (register char const *const s);
	
	int
	my_inet_aton (register char const *const host_name,
		      register struct in_addr *const addr);

	size_t
	xfgets (register char *const s,
		register size_t const size,
		register FILE *const ifile);
	
	size_t
	t_xfgets (register char *const buf,
	          register size_t const size,
	          register FILE *const xfile,
	          register unsigned int const timeout);
	
	size_t
	t_fread (register void *const buf,
	         register size_t const size,
	         register size_t const count,
	         register FILE *const xfile,
	         register unsigned int const timeout);
	
	size_t
	t_fwrite (register void const *const buf,
	          register size_t const size,
	          register size_t const count,
	          register FILE *const xfile,
	          register unsigned int const timeout);
	
	int
	t_connect (register int const s,
	           register struct sockaddr *const sp,
	           register socklen_t const size,
	           register unsigned int const timeout);
	
	void
	showline (register char *const line,
	          register size_t const len);
	
	int
	port_for (register char const *const service_name);
	
	const char *
	current_userid (void);
	
	typedef void (handler_func) (register int const signo);
	
	void
	set_signal_handler (register int const signo,
			    handler_func *const handler);

DESCRIPTION
	tbd

AUTHOR
	Ronald F. Guilmette <rfg@monkeys.com>

COPYRIGHT
	Copyright (c) 2002 Ronald F. Guilmette; All rights reserved.

	The author and copyright holder of this software specifically and
	explicitly DISALLOWS any and all uses of this software, or any
	compiled and/or derivative version thereof for any and all commercial
	purposes without the express written permission of the copyright
	holder to the contrary.

	Requests for permission to use this software, or any compiled and/or
	derivative work thereof, either in a commercial context or for com-
	mercial gain or for any commercial purpose should be e-mailed to the
	copyright holder at the e-mail address given above.

DISCLAIMER
	THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
	TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
	PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
	OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
	LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
	USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
	AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
	OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
	OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
	SUCH DAMAGE.
*/

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>  /* For MAXDNAME */
#include <netdb.h>
#include <pwd.h>
#include <fcntl.h>

#include "common.h"
#include "errors.h"
#include "utils.h"

typedef struct in_addr in_addr;
typedef struct sockaddr_in sockaddr_in;
typedef struct sockaddr sockaddr;
typedef struct sigaction Sigaction;
typedef struct linger linger;
typedef struct passwd passwd;
typedef struct timeval timeval;

enum { line_max = 1 << 10 };

void *
xmalloc (register unsigned long const size)
{
  register void *const result = (char *) malloc (size);

  if (!result)
    fatal ("Memory exhausted");
  return result;
}

char *
xstrdup (register char const *const s)
{
  register unsigned long const len = strlen (s) + 1;
  register char *const result = (char *) xmalloc (len);

  return (char *) memcpy (result, s, len);
}

void
xfree (register void *const ptr)
{
  free (ptr);
}

Bool
all_digits (register char const *const s)
{
  register size_t k;

  k = strspn (s, "0123456789");
  return (s[k] == '\0') ? True : False;
}

int
my_inet_aton (register char const *const host_name,
	      register struct in_addr *const addr)
{
  register int i;
  register char const *p;
  register int got_digits = 0;
  register unsigned num_num = 0;
  unsigned number[4] = { 0, 0, 0, 0 };
  register unsigned long a = 0;

  for (p = host_name; *p; p++)
    {
      if (*p == '.')
	{
	  if (!got_digits)
	    return 0;
	  got_digits = 0;
	  if (++num_num > 3)
	    return 0;
	}
      else if (!isdigit(*p))
	return 0;
      else
	{
          got_digits = 1;
          number[num_num] = (number[num_num] * 10) + (*p - '0');
	}
    }
  if (num_num != 3)
    return 0;
  for (i = 0; i < 4; i++)
    {
      if (number[i] > 255)
	return 0;
      a = (a << 8) | number[i];
    }
  addr->s_addr = htonl (a);
  return 1;
}

/* Like fgets, except this returns a size_t result, which will be 0 on error
   or EOF, or if `size' bytes have been read without encountering a newline.
   Otherwise the value returned is the number of bytes stored, including the
   terminating newline.

   Unlike fgets, this routine discards any nul bytes encountered on input.

   Unlike fgets, this routine DOES NOT store a nul byte after the input
   bytes stored in the string `s'.  */

size_t
xfgets (register char *const s,
	register size_t const size,
	register FILE *const ifile)
{
  register char const *const limit = s + size;
  register char *p;

  for (p = s;; )
    {
      register int const ch = getc (ifile);

      if (ch == EOF)
	return 0;
      else if (!ch)
	/* do nothing */;
      else if (p == limit)
	return 0;
      else
	{
#if 0
	  debug ("Char: \\%03o", (unsigned char) ch);
#endif
	  *p++ = ch;
	  if (ch == '\n')
	    return (size_t)(p - s);
	}
    }
}

static void
timeout_handler (register int const signo)
{
}

int
t_connect (register int const s,
           register struct sockaddr *const sp,
           register socklen_t const size,
           register unsigned int const timeout)
{
  auto Sigaction catch_timeout;
  auto Sigaction old_action;
  register int r;

  if (timeout)
    {
      catch_timeout.sa_handler = timeout_handler;
      sigemptyset (&catch_timeout.sa_mask);
      catch_timeout.sa_flags = 0;
      if (sigaction (SIGALRM, &catch_timeout, &old_action) == -1)
        fatal ("Error in sigaction: %s", strerror (errno));
      alarm (timeout);
    }
  r = connect (s, sp, size);
  if (timeout)
    {
      if (r == -1 && errno == EINTR)
        errno = ETIMEDOUT;
      alarm (0);
      if (sigaction (SIGALRM, &old_action, NULL) == -1)
        fatal ("Error in sigaction: %s", strerror (errno));
    }
  return r;
}

size_t
t_xfgets (register char *const buf,
          register size_t const size,
          register FILE *const xfile,
          register unsigned int const timeout)
{
  auto Sigaction catch_timeout;
  auto Sigaction old_action;
  register size_t r;

  if (timeout)
    {
      catch_timeout.sa_handler = timeout_handler;
      sigemptyset (&catch_timeout.sa_mask);
      catch_timeout.sa_flags = 0;
      if (sigaction (SIGALRM, &catch_timeout, &old_action) == -1)
        fatal ("Error in sigaction: %s", strerror (errno));
      alarm (timeout);
      errno = 0;
    }
  r = xfgets (buf, size, xfile);
  if (timeout)
    {
      if (errno == EINTR)
        errno = ETIMEDOUT;
      alarm (0);
      if (sigaction (SIGALRM, &old_action, NULL) == -1)
        fatal ("Error in sigaction: %s", strerror (errno));
    }
  return r;
}

size_t
t_fread (register void *const buf,
         register size_t const size,
         register size_t const count,
         register FILE *const xfile,
         register unsigned int const timeout)
{
  auto Sigaction catch_timeout;
  auto Sigaction old_action;
  register int r;

  if (timeout)
    {
      catch_timeout.sa_handler = timeout_handler;
      sigemptyset (&catch_timeout.sa_mask);
      catch_timeout.sa_flags = 0;
      if (sigaction (SIGALRM, &catch_timeout, &old_action) == -1)
        fatal ("Error in sigaction: %s", strerror (errno));
      alarm (timeout);
    }
  r = fread (buf, size, count, xfile);
  if (timeout)
    {
#if 0
      debug ("t_fread: %d %d %d", r, ferror (xfile), errno);
#endif
      if (ferror (xfile) && errno == EINTR)
        errno = ETIMEDOUT;
      alarm (0);
      if (sigaction (SIGALRM, &old_action, NULL) == -1)
        fatal ("Error in sigaction: %s", strerror (errno));
    }
  return r;
}

size_t
t_fwrite (register void const *const buf,
          register size_t const size,
          register size_t const count,
          register FILE *const xfile,
          register unsigned int const timeout)
{
  auto Sigaction catch_timeout;
  auto Sigaction old_action;
  register int r;

  if (timeout)
    {
      catch_timeout.sa_handler = timeout_handler;
      sigemptyset (&catch_timeout.sa_mask);
      catch_timeout.sa_flags = 0;
      if (sigaction (SIGALRM, &catch_timeout, &old_action) == -1)
        fatal ("Error in sigaction: %s", strerror (errno));
      alarm (timeout);
    }
  r = fwrite (buf, size, count, xfile);
  if (fflush (xfile) == EOF)
    r = 0;
  if (timeout)
    {
      if (ferror (xfile) && errno == EINTR)
        errno = ETIMEDOUT;
      alarm (0);
      if (sigaction (SIGALRM, &old_action, NULL) == -1)
        fatal ("Error in sigaction: %s", strerror (errno));
    }
  return r;
}

void
showline (register char *const line,
          register size_t const len)
{
  register int save1 = '\0';  /* Just initialized to prevent warnings.  */
  register int save2 = '\0';  /* Just initialized to prevent warnings.  */
  
  line[len] = '\0';
  
  if (len >= 2)
    {
      save2 = line[len-2];
      if (iscntrl (save2))
        line[len-2] = '\0';
    }
  if (len >= 1)
    {
      save1 = line[len-1];
      if (iscntrl (save1))
        line[len-1] = '\0';
    }
  info (">>> %s", line);
  if (len >= 1)
    line[len-1] = save1;
  if (len >= 2)
    line[len-2] = save2;
}

int
port_for (register char const *const service_name)
{
  register struct servent *sp;

  if ((sp = getservbyname (service_name, "tcp")) == NULL)
    fatal ("Error in getservbyname: %s", strerror (errno));
  return ntohs (sp->s_port);
}

char const *
current_userid (void)
{
  register int const uid = getuid ();
  register passwd const *const pw = getpwuid (uid);

  if (!pw)
    fatal ("No entry in /etc/passwd for UID %d", uid);
  return xstrdup (pw->pw_name);
}

void
set_signal_handler (register int const signo, handler_func *const handler)
{
  auto Sigaction sa;

  sa.sa_handler = handler;
  sigemptyset (&sa.sa_mask);
  sa.sa_flags = 0;
  if (sigaction (signo, &sa, NULL) == -1)
    fatal ("Error in sigaction: %s", strerror (errno));
}
